#!/usr/bin/env python

"""
This file is part of FuSpot.
Copyright (C) 2017 The Ohio State University

FuSpot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Please note that this software also uses pysam which is made open
source under the MIT license. Please see
<https://github.com/pysam-developers/pysam> for more information.

"""

"""
Description:	Given a list of bam files and a csv file of genomic coordinates, extracts all flanking reads from the bam file in exonic 
		regions out to a user specified distance (calculated as exonic distance) as well as genomic reads out to the same user 
		defined distance (calculated as genomic distance).
 Required Arguments: 
		@breakpoints_filename: name of the csv file containing a fusion breakpoint formatted a/t FuSpot manual
		@bamfiles: variable length list of bamfiles to extract reads from
"""


#############################################################################################
# Edit this constant to contain the location where you stored the downloaded exons.bed file #
#############################################################################################
EXON_BEDFILE = "exons_hg19.bed"



import argparse
import os
import time
import pysam
import math
from operator import itemgetter

def cprint(string,condition):
	if condition:
		print string

#####################
# EXTRACT FUNCTIONS #
#####################

#determine whether or not to use base10 binning based on user option, bam file size, and number of reads in a given batch. 
def use_base10_binning(bamfile, file_size, chrom, start, end, bin_option, verbose):
	#if user selected "force", use binning regardless of other parameters.
	if bin_option == "force":
		return True

	#holds return value
	use_binning = False

	#Constants for smart binning
	BATCH_SIZE_CUTOFF = 500		
	FILE_SIZE_CUTOFF = 1
	
	#2^30 is 1 GiB
	power = 30 + FILE_SIZE_CUTOFF - 1

	reads_in_batch = bamfile.count(chrom, start, end)
	cprint( "Bam file size: " + str(file_size), verbose)
	cprint( "Reads in this batch: " + str(reads_in_batch), verbose)
	
	#Check if force binning is selected or if the current file and batch size meet smart binning requirements.
	if (bin_option == "smart" and file_size > pow(2,power) and reads_in_batch >= BATCH_SIZE_CUTOFF):
		cprint( "Using base10 binning.", verbose)
		use_binning = True
	cprint("", verbose)

	return use_binning


def extract_reads_from_batch(batch, ref_bam, p_reads, do_binning, verbose):

	rounded_pos = 0
	count_reads_included = 0
	locations_processed = set()
	reads_out = [[],[]]

	i = 0
	for read in batch:
		process_this_read = read.qname not in p_reads
		if do_binning:
			rounded_pos = round(read.reference_start,-1)
			process_this_read = process_this_read and rounded_pos not in locations_processed
		if process_this_read:
			#cprint( "Processing Read: " + str(i), verbose)
			#append this read to reads_out
			reads_out[0].append('>' + read.qname)
			reads_out[0].append(read.seq)

			#find mate and append it to reads_out
			mate =  ref_bam.mate(read)
			reads_out[1].append('>' + mate.qname)
			reads_out[1].append(mate.seq)

			#account for this read in p_reads and, if necessary, locations_processed
			p_reads.add(read.qname)
			count_reads_included += 1
			if do_binning:
				locations_processed.add(rounded_pos)
		i+=1
	cprint("", verbose)
	cprint( "Reads processed: " + str(count_reads_included), verbose)
	cprint("", verbose)
	return reads_out



#name and open the output files
def extract(bamfile_in, breakpoints_file, exon_file, binning_type, verbose=False):

	cprint( "RUNNING ON NEW BAM FILE: " + bamfile_in, verbose)
	cprint("", verbose)
	#read in the regions file
	regions_file = open(breakpoints_file, 'r')

	regions=[]
	for line in regions_file:
		row = line.strip().split('\t')
		regions.append(row)
	regions_file.close()


	#set up a dictionary with keys as chromosome names and values as sorted array of locations
	#(exons.bed is already sorted so arrays will be sorted also)
	chroms={}
	#read in the exons from the bedfile
	bedfile = open(exon_file,'r')
	for exon in bedfile:
		exon = exon.strip()
		entries = exon.split('\t')
		bed_chrom = entries[0]
		bed_start = int(entries[1])
		bed_end = int(entries[2])
		bed_strand = entries[5]
		if bed_chrom in chroms:
			chroms[bed_chrom].append([bed_start,bed_end,bed_strand])
		else:
			chroms[bed_chrom]=[[bed_start,bed_end,bed_strand]]
	bedfile.close()

	#to keep track of read names that we have processed so that we do not extract reads more than once.
	reads_processed = set()

	#read name goes in first list, read sequence goes in second list
	reads_out_final= [[],[]]

	#loop to get reads from all the regions of breakpoints.csv
	for region in regions:
		chrom = region[0]
		breakpoint = int(region[1])
		radius = int(region[2])
		
		left_exon_frags=[]
		right_exon_frags=[]

		#find the exonic regions that we need to interrogate. 
		#Input "+" for strand. Doesn't matter since we check both directions anyway.
		left_exon_frags = find_exon(chrom,breakpoint,radius,"+","left",chroms)
		right_exon_frags = find_exon(chrom,breakpoint,radius,"+","right",chroms)
		cprint( "Number of left exon frags: " + str(len(left_exon_frags)), verbose )
		cprint( "Number of right exon frags: " + str(len(right_exon_frags)), verbose )

		#open a version of the bamfile to iterate over and a version to fetch mates from
		bamfile = pysam.Samfile(bamfile_in, 'rb')
		ref_bamfile = pysam.Samfile(bamfile_in, 'rb')

		#get bam file size
		file_size = os.path.getsize(bamfile_in)

		#loop over left exons
		for i,exon in enumerate(left_exon_frags):
			exon_start = exon[0]
			exon_end = exon[1]
			batch = None
			try:
				batch = bamfile.fetch(chrom, exon_start, exon_end)
			except(ValueError):
				chrom = chrom[3:]
				batch = bamfile.fetch(chrom, exon_start, exon_end)
			
		
			cprint( "Checking left exon: " + str(i), verbose)
			cprint( "Possible read locations with binning: " + str(math.ceil((exon_end - exon_start)/10)), verbose)
			start = time.time()
			#check user option and other parameters to determine whether to use binning.
			use_binning = use_base10_binning(bamfile, file_size, chrom, exon_start, exon_end, binning_type, verbose)

			exon_reads_out = extract_reads_from_batch(batch, ref_bamfile, reads_processed, use_binning, verbose)
			done = time.time()

			cprint("",verbose)
			elapsed = done - start

			m, s = divmod(elapsed, 60)
			h, m = divmod(m, 60)
			tstring =  "%d:%02d:%02d" % (h, m, s)
			cprint(tstring, verbose)
			cprint("",verbose)

			#append the extracted reads to the final list of reads
			for i,entry in enumerate(exon_reads_out):
				reads_out_final[i]+=entry

		#get genomic reads
		#	-take max of end of last left exon and bp-radius
		#	-take min of start of first right exon and bp+radius
		#	-if second - first == 1 then skip.
		last_exon_end = left_exon_frags[-1][1]
		genomic_start = breakpoint - radius
		central_start = max(last_exon_end, genomic_start)

		first_exon_start = right_exon_frags[0][0]
		genomic_end = breakpoint + radius
		central_end = min(first_exon_start,genomic_end)
		
		if ( (central_end - central_start) > 1 ):
			batch = None
			try:
				batch = bamfile.fetch(chrom, central_start, central_end)
			except(ValueError):
				chrom = chrom[3:]
				batch = bamfile.fetch(chrom, central_start, central_end)

			start = time.time()

			cprint( "Checking genomic region: " + chrom + " " + str(central_start) + "-" + str(central_end), verbose)
			cprint( "Possible read locations with binning: " + str(math.ceil((exon_end - exon_start)/10)), verbose)
			use_binning = use_base10_binning(bamfile, file_size, chrom, central_start, central_end, binning_type, verbose)

			local_reads_out = extract_reads_from_batch(batch, ref_bamfile, reads_processed, use_binning, verbose)
			done = time.time()

			cprint("",verbose)
			elapsed = done - start

			m, s = divmod(elapsed, 60)
			h, m = divmod(m, 60)
			tstring =  "%d:%02d:%02d" % (h, m, s)
			cprint(tstring, verbose)
			cprint("",verbose)

			#append the extracted reads to the final list of reads
			for i,entry in enumerate(local_reads_out):
				reads_out_final[i]+=entry

		#loop over right exons
		for i,exon in enumerate(right_exon_frags):
			exon_start = exon[0]
			exon_end = exon[1]
			batch = None
			try:
				batch = bamfile.fetch(chrom, exon_start, exon_end)
			except(ValueError):
				chrom = chrom[3:]
				batch = bamfile.fetch(chrom, exon_start, exon_end)

			start = time.time()
			cprint( "Checking left exon: " + str(i), verbose)

			cprint( "Possible read locations with binning: " + str(math.ceil((exon_end - exon_start)/10)), verbose)

			use_binning = use_base10_binning(bamfile, file_size, chrom, exon_start, exon_end, binning_type, verbose)
			exon_reads_out = extract_reads_from_batch(batch, ref_bamfile, reads_processed, use_binning, verbose)

			done = time.time()

			cprint("",verbose)
			elapsed = done - start

			m, s = divmod(elapsed, 60)
			h, m = divmod(m, 60)
			tstring =  "%d:%02d:%02d" % (h, m, s)
			cprint(tstring, verbose)
			cprint("",verbose)

			#append the extracted reads to the final list of reads
			for i,entry in enumerate(exon_reads_out):
				reads_out_final[i]+=entry
		
		bamfile.close()
		ref_bamfile.close()

	cprint( "Reads returned: " + str(len(reads_out_final[0])), verbose)
	return reads_out_final

#######################
# FIND-EXON FUNCTIONS #
#######################

# "extract" the number of bases specifed by bases_to_extract from exon.
# if bases_to_extract is longer than exon, it will return the difference
# appends the "sequences" as [start,end] to seqs
def get_exon_seq_positions(exon,start,direction,bases_to_extract,seqs):

    seq_start = 0
    seq_end = 0
    bases_remaining = 0
    exon_start=exon[0]
    exon_end=exon[1]

    #make sure that the specified start position is inside the exon
    if not (start >= exon_start and start <= exon_end):
        raise ValueError("Start position is not inside given exon")
    #if direction is left, look from end to start
    if direction == "left":
        if start - exon_start +1 >= bases_to_extract:
            seq_start = start - bases_to_extract + 1
            seq_end = start
            bases_remaining = 0
        else:
            seq_start = exon_start
            seq_end = start
            bases_remaining = bases_to_extract - (seq_end - seq_start + 1)
            #print seq_start
            #print seq_end
            #print bases_remaining
    elif direction == "right":
        if exon_end - start >= bases_to_extract - 1 :
            seq_start = start
            seq_end = seq_start + bases_to_extract - 1 
            bases_remaining = 0
        else:
            seq_start = start
            seq_end = exon_end
            bases_remaining = bases_to_extract - (seq_end - seq_start + 1)
    else:
        raise ValueError("Please pick a valid fusion side")
    #print bases_remaining
    seqs.append([seq_start,seq_end])
    return bases_remaining

# Given a list of exons, one genomic coordinate, and a direction to search, identifies the nearest (or current)
# exon and provides a list of the genomic ranges corresponding to exonic sequences
# such that the ranges can be used to query hg19 to provide exactly the number of 
# bases specified by @radius.
#
# Parameters:
# chromosome: chromosome of genomic coordinate
# position: base pair position of genomic coordinate
# radius: desired length of exonic sequence
# side: direction to look for sequences
# chroms: a dictionary with keys = chromosome names, values = list of exons, 
#         exons = [start, end, strand]

def find_exon(chromosome,position,radius,strand,side,chroms):
    #CONSTANTS
    IN_EXON_STR = "IS_IN_EXON"
    EXON_BOUND_STR = "IS_EXON_BOUNDARY"
    edge_distance = 3	#if the breakpoint and exon edge are within this distance, breakpoint considered on edge

#process which side we need to check
    d = {}
    d["left"]="right"
    d["right"]="left"

    if not side in d:
        raise ValueError("Please pick a valid fusion side (right or left.)")

    side_to_check = side
    if (strand == "-"):
        side_to_check = d[side_to_check]


    #find the closest exon in the given direction
    #if left, as soon as we find one greater, take one step back.
    #if right, as soon as we find one greater, stop.
    is_exon_start_boundary = False
    is_exon_end_boundary = False
    is_in_exon = False
    #Can exons jump to the next chromosome? Do we need to check the next chr in the dict if we don't find an exon?
    usrstart = position
    desired_exon_index = -2
    check_end_of_exon_with_index = 0
    for i,exon in enumerate(chroms[chromosome]):
        exonstart = exon[0]
        if (usrstart <= exonstart or abs(usrstart - exonstart) < edge_distance):
            check_end_of_exon_with_index = i-1
            if side_to_check == "left":
                desired_exon_index = i-1
                if(abs(usrstart - exonstart) < edge_distance):
                    is_exon_start_boundary = True
                    desired_exon_index = i
            #if looking right
            else:
                #assume this position is not in the exon
                desired_exon_index = i
                if(abs(usrstart - exonstart) < edge_distance):
                    is_exon_start_boundary = True
        
            # "Desired exon index " + str(desired_exon_index)
            break
    #print i

    #check if the position is before the first exon in the exons for that chrom
    before_all_exons = False
    after_all_exons = False
    if(check_end_of_exon_with_index == -1 and not is_exon_start_boundary):
        desired_exon_index = 0
        before_all_exons = True
    #after all exons?
    if(desired_exon_index == -2):
        check_end_of_exon_with_index = i
        if side_to_check == "left":
            desired_exon_index = i

    #check if the given position was inside an exon (or on a boundary for the left case)
    exon_to_check = chroms[chromosome][check_end_of_exon_with_index]
    check_end = exon_to_check[1]
    if (usrstart <= check_end and not (is_exon_start_boundary or before_all_exons or after_all_exons)):
        #print "usrstart and check_end"
        #print usrstart
        #print check_end
        if(usrstart == check_end):
            is_exon_end_boundary = True
            #print "IS EXON BOUNDARY"
        else:
            #print "Pos: " + str(check_end)
            #print "Index: " + str(check_end_of_exon_with_index)
            #print exon_to_check
            is_in_exon = True
            #if it is in the exon, we need the one at the previous index.
            if (side_to_check == "right"):
                desired_exon_index-=1
    #print "exon..."
    #print chroms[chromosome][desired_exon_index]

    #Out list of exon sequences to wget from the das server
    exon_seqs=[]

#    if (side_to_check == "left"):
#        if(before_all_exons):
#            raise ValueError("We need left side but this position is before all exons in this chr.")
#    elif(side_to_check == "right"):
#        if(after_all_exons):
#            raise ValueError("We need right side but this position is after all exons in this chr.")

    desired_chrom = chroms[chromosome]
    desired_exon = desired_chrom[desired_exon_index]
    out_chr = chromosome

    #print desired_exon

    #figure where to start based on whether it is inside or outside exon
    start = 0
    is_exon_boundary = is_exon_start_boundary or is_exon_end_boundary
    if(is_exon_boundary or is_in_exon):
        if(is_exon_start_boundary and side_to_check == "left"):
            desired_exon_index-=1
            desired_exon = desired_chrom[desired_exon_index]
            start = desired_exon[1]
        elif(is_exon_end_boundary and side_to_check == "right"):
            #desired_exon_index+=1
            desired_exon = desired_chrom[desired_exon_index]
            start = desired_exon[0]
        else:
            start = usrstart
    else:
        if (side_to_check == "left"):
            start = desired_exon[1]
        else:
            start = desired_exon[0]
    bases_remaining = radius

    #loop to find exon sequences
    #print "Start position: " + str(start)
    bases_remaining = get_exon_seq_positions(desired_exon,start,side_to_check,bases_remaining,exon_seqs)
    while(bases_remaining > 0):
        #print "bases remaining: " + str(bases_remaining)
        if(side_to_check == "left"):
            desired_exon_index-=1
        else:
            desired_exon_index+=1

        desired_exon = desired_chrom[desired_exon_index]
        if (side_to_check == "left"):
            start = desired_exon[1]
        else:
            start = desired_exon[0]

        bases_remaining =  get_exon_seq_positions(desired_exon,start,side_to_check,bases_remaining,exon_seqs) 
    
    exon_seqs=sorted(exon_seqs, key=itemgetter(0))
    #print "Exon sequences: "
    #print exon_seqs
    return exon_seqs



########
# MAIN #
########

DESCRIPTION = "Given a list of bam files and a csv file of genomic coordinates, extracts all flanking reads from the bam file in exonic regions out to a user specified distance (calculated as exonic distance) as well as genomic reads out to the same user defined distance (calculated as genomic distance)."

BREAKPOINT_HELP = "CSV file with genomic coordinates of fusion breakpoint. See README for format."
BAMFILE_HELP = "One or more bamfiles to extract reads from."
PREFIX_HELP = "Prefix for output fasta filenames"
SMART_BINNING_HELP = "When dealing with large bam files, extraction may take a considerable amount of time, especially when many duplicate reads are present. Activate this option to speed up extraction by taking only one read from every window of 10 base pairs for file sizes over 1gb and for regions containing > 500 reads. Note: This will also potentially skip mates that may align over the fuison point. Leave this option deactivated to see ALL supporting reads."
FORCE_BINNING_HELP = "Force binning will bin reads from all files regardless of size or number of reads in region. Note: Smart Binning and Force Binning are mutually exclusive."
VERBOSE_HELP = "Print progress messages."
EXON_BED_HELP = "Bed file of exons. Must be sorted by position."

parser = argparse.ArgumentParser(description = DESCRIPTION)
parser.add_argument("breakpoints_filename", type = str, help = BREAKPOINT_HELP)
parser.add_argument("bamfiles", type = str, nargs='+', help = BAMFILE_HELP)
parser.add_argument("-x", "--prefix", type = str, default = "reads", help = PREFIX_HELP)
parser.add_argument("-sbin","--smart_binning", action = "store_true", help = SMART_BINNING_HELP)
parser.add_argument("-fbin","--force_binning", action = "store_true", help = FORCE_BINNING_HELP)
parser.add_argument("-v","--verbose", action = "store_true", help = VERBOSE_HELP)
parser.add_argument("-b","--exon_bed", type = str, default = EXON_BEDFILE, help = EXON_BED_HELP)
args= parser.parse_args()

start = time.time()

binning_type = ""
if (args.smart_binning and args.force_binning):
	raise ValueError("Smart Binning and Force Binning are mutually exclusive.")
elif (args.smart_binning):
	binning_type = "smart"
elif (args.force_binning):
	binning_type = "force"

# get reads from all bam files
all_reads = [[],[]]
for bamfile in args.bamfiles:
	#print bamfile
	reads = extract(bamfile, args.breakpoints_filename, args.exon_bed,binning_type,verbose=args.verbose)
	all_reads[0]+=reads[0]
	all_reads[1]+=reads[1] 

# cat all the read files together, maintaining pair order.

fa_name1 = args.prefix + "_R1.fasta"
fa_name2 = args.prefix + "_R2.fasta"

 
if os.path.isfile(fa_name1):
	os.remove(fa_name1)
	#print "Replacing: " + fa_name1

fa_file1 = open(fa_name1, 'w')
for line in all_reads[0]:
	print >> fa_file1, line
fa_file1.close()

if os.path.isfile(fa_name2):
	os.remove(fa_name2)
	#print "Replacing: " + fa_name2

fa_file2 = open(fa_name2, 'w')
for line in all_reads[1]:
	print >> fa_file2, line
fa_file2.close()

done = time.time()
elapsed = done - start

m, s = divmod(elapsed, 60)
h, m = divmod(m, 60)
print "Extraction time: %d:%02d:%02d" % (h, m, s)
